import pytest

from tranquilizer.decorator import tranquilize, publish
from tranquilizer.types import File, TextFile, Image, NDArray
from tranquilizer.types import dt_factory, list_factory
from tranquilizer.types import is_container, is_list_subclass
from tranquilizer.types import type_mapper
from werkzeug.datastructures import FileStorage
import typing
import numpy as np
import datetime
import PIL.Image
import io
from os.path import join, dirname
import pandas as pd

def test_is_container():
    l = [0,1,2]
    d = {'a':0, 'b':1}
    
    for c in l,d:
        assert is_container(type(c)) == True
    
    f = 3.1415
    s = 'tranquilizer'
    i = 42

    for c in f,s,i:
        assert is_container(type(c)) == False
    
    assert is_container(typing.List) == True

    # numpy arrays are sent as a single chunk of binary
    # data
    arr = np.array(l)
    assert is_container(type(arr)) == False

def test_is_list_subclass():
    L = typing.List
    Li = typing.List[int]
    
    assert is_list_subclass(list) == True
    assert is_list_subclass(L) == True
    assert is_list_subclass(Li) == True

    assert is_list_subclass(int) == False
    assert is_list_subclass(str) == False

def test_list_factory():
    integers = list_factory(int)
    assert integers('0') is 0
    with pytest.raises(ValueError):
        integers('a')
        integers('3.14')

    floats = list_factory(float)
    assert np.isclose(floats('3.14'), 3.14)
    with pytest.raises(ValueError):
        floats('a')
    
    dates = list_factory(dt_factory(datetime.date))
    assert dates('2018-1-1 4 PM') == datetime.date(2018,1,1)

    assert not (integers == floats)

def test_dt_factory():
    d = dt_factory(datetime.date)
    assert d('2018-1-1 4 PM') == datetime.date(2018,1,1)

    dt = dt_factory(datetime.datetime)
    assert dt('2018-1-1 4 PM') == datetime.datetime(2018,1,1,16,0,0)

    ts = dt_factory(pd.Timestamp)
    assert ts('2018-1-1 4 PM') == datetime.datetime(2018,1,1,16,0,0)


def test_type_mapper_builtins():
    assert int is type_mapper(int)
    assert float is type_mapper(float)
    assert bool is type_mapper(bool)
    assert str is type_mapper(str)

def test_type_mapper_date():
    dm = type_mapper(datetime.date)
    df = dt_factory(datetime.date)
    assert dm('2018-1-1 4 PM') == df('2018-1-1 4 PM') == datetime.date(2018,1,1)

    dtm = type_mapper(datetime.datetime)
    dtf = dt_factory(datetime.datetime)
    assert dtm('2018-1-1 4 PM') == dtf('2018-1-1 4 PM') == datetime.datetime(2018,1,1,16,0,0)

def test_type_mapper_list():
    l1 = list_factory(int) 
    l2 = type_mapper(typing.List[int])
    assert l1('0') == l2('0') == 0

    l1 = list_factory(float) 
    l2 = type_mapper(typing.List[float])
    assert l1('3.14') == l2('3.14') == 3.14

    l1 = list_factory(str) 
    l2 = type_mapper(list)
    assert l1('a') == l2('a') == 'a'

    l1 = list_factory(str) 
    l2 = type_mapper(typing.List)
    assert l1('a') == l2('a') == 'a'

    l1 = list_factory(str) 
    l2 = type_mapper(typing.List[str])
    assert l1('a') == l2('a') == 'a'

    l1 = list_factory(bool) 
    l2 = type_mapper(typing.List[bool])
    assert l1('True') == l2('True') == True

    l1 = list_factory(dt_factory(datetime.date))
    l2 = type_mapper(typing.List[datetime.date])
    assert l1('2018-1-1 4 PM') == l2('2018-1-1 4 PM') == datetime.date(2018,1,1)

def test_type_mapper_files():
    assert type_mapper(typing.BinaryIO) is File
    assert type_mapper(typing.TextIO) is TextFile
    assert type_mapper(PIL.Image.Image) is Image
    assert type_mapper(np.ndarray) is NDArray

def test_binary_file():
    here = dirname(__file__)
    wf = FileStorage(filename=join(here,'file'))
    assert File(wf) is wf
    attrs = ['__location__', '__schema__', '__description__']
    assert all([hasattr(File, a) for a in attrs])

def test_text_file():
    here = dirname(__file__)
    wf = FileStorage(filename=join(here,'file'))
    tf = TextFile(wf)
    assert isinstance(tf, io.StringIO)
    attrs = ['__location__', '__schema__', '__description__']
    assert all([hasattr(TextFile, a) for a in attrs])

def test_array_file():
    here = dirname(__file__)
    wf = FileStorage(filename=join(here,'arr.npy'))
    nf = NDArray(wf.filename)
    assert isinstance(nf, np.ndarray)
    attrs = ['__location__', '__schema__', '__description__']
    assert all([hasattr(NDArray, a) for a in attrs])

def test_image_file():
    here = dirname(__file__)
    wf = FileStorage(filename=join(here,'kitty.jpg'))
    img = Image(wf.filename)
    assert isinstance(img, PIL.Image.Image)
    attrs = ['__location__', '__schema__', '__description__']
    assert all([hasattr(TextFile, a) for a in attrs])
    


'''execute the script file'''
import ast
from runpy import run_path
from unittest.mock import MagicMock
from os.path import dirname, join, basename
import tempfile

def _is_tranquilized(decorator):
    return hasattr(decorator, 'id') and (decorator.id in ('tranquilize', 'publish'))

def _is_decorated(item):
    if isinstance(item, ast.FunctionDef):
        return any([_is_tranquilized(d.func) for d in item.decorator_list])


class BaseHandler(object):

    def parse(self): # pragma no cover
        raise NotImplementedError()

    @property
    def tranquilized_functions(self):
        self.parse()

        tranquilized = [f.name for f in self.nodes.body if _is_decorated(f)]
        functions = [self.module[f] for f in tranquilized]

        return functions


class ScriptHandler(BaseHandler):
    def __init__(self, fn):
        self.fn = fn

    def parse(self):
        with open(self.fn, 'r') as f:
            source = f.read()

        self.nodes = ast.parse(source, self.fn)
        self.module = run_path(self.fn)


class NotebookHandler(BaseHandler):
    def __init__(self, fn):
        self.fn = fn

    def parse(self):
        from nbconvert import ScriptExporter

        exporter = ScriptExporter()
        source, _ = exporter.from_filename(self.fn)

        self.nodes = ast.parse(source, self.fn)

        with tempfile.NamedTemporaryFile(mode='w', dir=dirname(self.fn), delete=True) as tmp:
            tmp.write(source)
            tmp.flush()
            self.module = run_path(tmp.name, init_globals={'get_ipython':MagicMock()})




'''construct flask-restx resources'''
from flask import jsonify
from flask_restx import Resource, reqparse
from collections.abc import Mapping, Sequence
from typing import List

from .types import is_container, type_mapper


def make_parser(func_spec, location='args', compat=False):
    '''Create RequestParser from annotated function arguments

    arguments without default values are flagged as required

    Parameters
    ----------
    :param func_spec: The decorated functions spec
    :param location: expected location of arguments in request
                     'args', or 'body'
    :param compat: If compatibility with anaconda-enterprise-web-publisher
                   is required. (default: False)
    '''

    parser = reqparse.RequestParser(bundle_errors=True)
    for argument, spec in func_spec['args'].items():
        if compat and spec.get('annotation', None) is None:
            # un-typed arguments are returned to the
            # function as a list of strings by @publish
            action = 'append'
            _type = List[str]
        else:
            # @tranquilize assumes untyped
            # arguments are strings
            _type = spec.get('annotation', str)
            action = 'store'

        _default = spec.get('default', None)
        _type = type_mapper(_type)
        doc = func_spec['param_docs'].get(argument, None)

        # Files (e.g., images) arrive in a different
        # Flask.Request location. The last value in
        # the tuple takes precedence.
        try:
            _loc = getattr(_type, '__location__')
            _location = (location, _loc)
        except AttributeError:
            _location = location

        if is_container(_type):
            action = 'append'

        parser.add_argument(argument, type=_type,
                            default=_default,
                            required=(not ('default' in spec)),
                            location=_location,
                            action=action,
                            help=doc)

    return parser


def make_resource(func, api):
    if func._methods:
        return _make_resources(func, api)
    else:
        return _make_resource(func, api, func._method)


def _make_resources(func, api):
    '''Provide compatibility with web-publisher'''
    resources = {}
    for m in func._methods:
        resources[m] = getattr(_make_resource(func, api, m), m)

    Tranquilized = type('Tranquilized', (Resource,), resources)

    return Tranquilized


def _make_resource(func, api, method):
    location = 'form' if method in ['put','post'] else 'args'
    compat = True if (func._methods is not None) else False
    parser = make_parser(func._spec, location=location, compat=compat)

    @api.expect(parser, validate=True, strict=True)
    def _method(self):
        req = parser.parse_args()
        output = func(**req)
        return jsonify(output)

    error_docs = func._spec['error_docs']
    if error_docs:
        _method = api.doc(responses=error_docs)(_method)

    _method.__doc__ = func._spec['docstring']

    Tranquilized = type('Tranquilized', (Resource,), {method:_method})

    return Tranquilized




from tranquilizer.decorator import tranquilize, publish
from tranquilizer.decorator import _prepare, _prepare_arg, _prepare_arg_docs
from tranquilizer.decorator import _prepare_error_docs
from inspect import signature
import datetime
import typing
import PIL.Image
import numpy

def test_attributes():

    def _func():
        return 0

    decorated = tranquilize()(_func)
    assert hasattr(decorated, '_spec')
    assert hasattr(decorated, '_method')
    assert hasattr(decorated, '_methods')
    assert decorated._methods is None


def test_publish_attributes():

    def _func():
        return 0

    decorated = publish()(_func)
    assert hasattr(decorated, '_spec')
    assert hasattr(decorated, '_method')
    assert hasattr(decorated, '_methods')
    assert decorated._method is None


def test_method():

    # separate functions are used for
    # get and post. Calling the decorator
    # a second time updates the original
    # function
    def _funcg():
        return 0

    get = tranquilize(method='GET')(_funcg)
    assert get._method == 'get'

    def _funcp():
        return 0

    post = tranquilize(method='PosT')(_funcp)
    assert post._method == 'post'

    def _funcput():
        return 0

    post = tranquilize(method='pUt')(_funcput)
    assert post._method == 'put'


def test_methods():

    # separate functions are used for
    # get and post. Calling the decorator
    # a second time updates the original
    # function
    def _funcg():
        return 0

    get = publish(methods=['GET'])(_funcg)
    assert get._methods == ['get']

    def _funcp():
        return 0

    post = publish(methods=['PosT'])(_funcp)
    assert post._methods == ['post']

    def _funcpg():
        return 0

    post_get = publish(methods=['GET', 'PosT'])(_funcpg)
    assert post_get._methods == ['get', 'post']


def test_prepare():
    def _func(arg: float):
        '''docstring

        :param arg: number
        :raises ValueError: not a number'''
        return arg
    
    spec = _prepare(_func)   

    assert isinstance(spec, dict)
    assert spec.keys() == set(['name','docstring','args', 
                               'param_docs','error_docs'])
    assert spec['name'] == '_func'
    assert spec['docstring'] == 'docstring\n\n        '

def test_prepare_no_docstring():
    def _empty(arg: float):
        pass

    spec = _prepare(_empty)
    assert spec['param_docs'] == {}
    assert spec['error_docs'] == {}
    assert spec['docstring'] == ''


def test_prepare_args():
    def _func(
        s: str,
        i: int,
        f: float,
        b: bool,
        d: datetime.date,
        dt: datetime.datetime,
        l: list,
        L: typing.List,
        Ls: typing.List[str],
        Li: typing.List[int],
        Lf: typing.List[float],
        Lb: typing.List[bool],
        Ld: typing.List[datetime.date],
        Ldt: typing.List[datetime.datetime],
        fnb: typing.BinaryIO,
        fnt: typing.TextIO,
        img: PIL.Image.Image,
        arr: numpy.ndarray,
        untyped,
        untyped_default = None,
        typed_default: str = 'python'
        
    ):
        pass

    sig = signature(_func)
    assert _prepare_arg(sig.parameters['s']) == {'name':'s','type':'str','annotation':str}
    assert _prepare_arg(sig.parameters['i']) == {'name':'i','type':'int','annotation':int}
    assert _prepare_arg(sig.parameters['f']) == {'name':'f','type':'float','annotation':float}
    assert _prepare_arg(sig.parameters['b']) == {'name':'b','type':'bool','annotation':bool}
    assert _prepare_arg(sig.parameters['d']) == {'name':'d','type':'date','annotation':datetime.date}
    assert _prepare_arg(sig.parameters['dt']) == {'name':'dt','type':'datetime','annotation':datetime.datetime}
    assert _prepare_arg(sig.parameters['l']) == {'name':'l','type':'list','annotation':list}
    assert _prepare_arg(sig.parameters['L']) == {'name':'L','type':'List','annotation':typing.List}
    assert _prepare_arg(sig.parameters['Ls']) == {'name':'Ls','type':'List','annotation':typing.List[str]}
    assert _prepare_arg(sig.parameters['Li']) == {'name':'Li','type':'List','annotation':typing.List[int]}
    assert _prepare_arg(sig.parameters['Lf']) == {'name':'Lf','type':'List','annotation':typing.List[float]}
    assert _prepare_arg(sig.parameters['Lb']) == {'name':'Lb','type':'List','annotation':typing.List[bool]}
    assert _prepare_arg(sig.parameters['Ld']) == {'name':'Ld','type':'List','annotation':typing.List[datetime.date]}
    assert _prepare_arg(sig.parameters['Ldt']) == {'name':'Ldt','type':'List','annotation':typing.List[datetime.datetime]}
    assert _prepare_arg(sig.parameters['fnb']) == {'name':'fnb','type':'BinaryIO','annotation':typing.BinaryIO}
    assert _prepare_arg(sig.parameters['fnt']) == {'name':'fnt','type':'TextIO','annotation':typing.TextIO}
    assert _prepare_arg(sig.parameters['img']) == {'name':'img','type':'Image','annotation':PIL.Image.Image}
    assert _prepare_arg(sig.parameters['arr']) == {'name':'arr','type':'ndarray','annotation':numpy.ndarray}
    assert _prepare_arg(sig.parameters['untyped']) == {'name':'untyped'}
    assert _prepare_arg(sig.parameters['untyped_default']) == {'name':'untyped_default', 'default': None}
    assert _prepare_arg(sig.parameters['typed_default']) == {'name':'typed_default', 'type':'str', 'annotation':str, 'default': 'python'}

def test_prepare_arg_docs():
        doc = '''docstring

        docstring

        :param arg1: number
        :param arg2: string
        :raises ValueError: not a number'''

        param_docs, remainder = _prepare_arg_docs(doc)
        
        assert param_docs == {'arg1': 'number', 'arg2': 'string'}
        assert remainder == 'docstring\n\n        docstring\n\n        :raises ValueError: not a number'

def test_prepare_arg_doc_noargs():
        doc = '''docstring

        docstring

        :raises ValueError: not a number'''

        param_docs, remainder = _prepare_arg_docs(doc)

        assert param_docs == {}
        assert remainder == doc

def test_prepare_error_docs():
        doc = '''docstring

        docstring

        :param arg1: number
        :param arg2: string
        :raises ValueError: not a number'''

        error_docs, remainder = _prepare_error_docs(doc)
        
        assert error_docs == {500:'ValueError:not a number'}
        assert remainder == 'docstring\n\n        docstring\n\n        :param arg1: number\n        :param arg2: string\n        '

def test_prepare_error_noerror():
        doc = '''docstring

        docstring

        :param arg1: number
        :param arg2: string'''

        error_docs, remainder = _prepare_error_docs(doc)
        
        assert error_docs is None
        assert remainder == doc


from tranquilizer.handler import ScriptHandler, NotebookHandler
from os.path import dirname, join

def test_script_handler():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.py')
    script = ScriptHandler(fn)
    assert script.fn == fn

def test_script_parser():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.py')
    script = ScriptHandler(fn)
    script.parse()
    assert [hasattr(script, attr) for attr in ['modules','nodes']]

def test_nb_handler():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.ipynb')
    nb = NotebookHandler(fn)
    assert nb.fn == fn

def test_nb_parser():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.ipynb')
    nb = NotebookHandler(fn)
    nb.parse()
    assert [hasattr(nb, attr) for attr in ['modules','nodes']]

def test_tranquilized_script():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.py')
    script = ScriptHandler(fn)
    script.parse()
    funcs = script.tranquilized_functions
    assert len(funcs) == 1

def test_tranquilized_nb():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.ipynb')
    nb = NotebookHandler(fn)
    nb.parse()
    funcs = nb.tranquilized_functions
    assert len(funcs) == 1

def test_published_script():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop_publish.py')
    script = ScriptHandler(fn)
    script.parse()
    funcs = script.tranquilized_functions
    assert len(funcs) == 1


from tranquilizer.application import make_app
from tranquilizer.handler import ScriptHandler
from werkzeug.middleware.proxy_fix import ProxyFix
from os.path import dirname, join

def test_app_function():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.py')
    script = ScriptHandler(fn)
    script.parse()
    funcs = script.tranquilized_functions

    app = make_app(funcs, 'cheese')
    assert len(app.blueprints) == 1

def test_app_function():
    here = dirname(__file__)
    fn = join(here, 'two_funcs.py')
    script = ScriptHandler(fn)
    script.parse()
    funcs = script.tranquilized_functions

    app = make_app(funcs, 'two_funcs')
    assert len(app.blueprints) == 1

def test_proxy_fix():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.py')
    script = ScriptHandler(fn)
    script.parse()
    funcs = script.tranquilized_functions

    app = make_app(funcs, 'cheese')
    assert isinstance(app.wsgi_app, ProxyFix)

def test_content_length():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.py')
    script = ScriptHandler(fn)
    script.parse()
    funcs = script.tranquilized_functions

    app = make_app(funcs, 'cheese', max_content_length=1024)
    assert app.config['MAX_CONTENT_LENGTH'] == 1024


import sys
from os.path import dirname, basename
from argparse import ArgumentParser

from .application import make_app
from .handler import ScriptHandler, NotebookHandler
from .__init__ import __version__


class UnsupportedFileType(Exception):
    pass

def cli():
    parser = ArgumentParser(prog="tranquilizer",
                            description="Put your functions to REST")
    parser.add_argument('filename', help='Script file with tranquilized functions')
    parser.add_argument('--name', help='Name of the REST API to use in Swagger')
    parser.add_argument('--max_content_length', type=int,
                        help='Maximum size of request in bytes for all endpoints')

    parser.add_argument('--port', '--anaconda-project-port', action='store', default=8086, type=int,
                        help='Port to listen on')
    parser.add_argument('--address', '--anaconda-project-address',
                        action='store',
                        default='0.0.0.0',
                        help='IP address the application should listen on.')
    parser.add_argument('--prefix','--anaconda-project-url-prefix', action='store', default='',
                        help='Prefix in front of urls')

    parser.add_argument('--debug', action='store_true', default=False,
                        help='Run API with debug output.')
    parser.add_argument('--version', action='version',
                        version='%(prog)s {version}'.format(version=__version__))

    # these anaconda-project arguments are ignored by Tranquilizer
    parser.add_argument('--anaconda-project-host', action='append', default=[],
                        help='Hostname to allow in requests')
    parser.add_argument('--anaconda-project-iframe-hosts',
                        action='append',
                        help='Space-separated hosts which can embed us in an iframe per our Content-Security-Policy')
    parser.add_argument('--anaconda-project-no-browser', action='store_true',
                        default=False,
                        help='Disable opening in a browser')
    parser.add_argument('--anaconda-project-use-xheaders',
                        action='store_true',
                        default=False,
                        help='Trust X-headers from reverse proxy')

    return parser


def main(args):
    sys.path.append(dirname(args.filename))

    extension = args.filename.split('.')[-1]
    if extension == 'py':
        source = ScriptHandler(args.filename)
    elif extension == 'ipynb':
        try:
            import nbconvert
        except ImportError as e: # pragma no cover
            raise ImportError("Please install nbconvert to serve Jupyter Notebooks.") from e

        source = NotebookHandler(args.filename)
    else:
        raise UnsupportedFileType('{} is not a script (.py) or notebook (.ipynb)'.format(args.filename))

    name = args.name if args.name else basename(args.filename)
    app = make_app(source.tranquilized_functions, name=name, prefix=args.prefix,
                   max_content_length=args.max_content_length)

    return app

def run():
    args = cli().parse_args()
    app = main(args)
    app.run(host=args.address, port=args.port,
            debug=args.debug)


from tranquilizer.main import cli, main, UnsupportedFileType
from tranquilizer.__init__ import __version__
from os.path import dirname, join
import pytest

def test_py():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.py')
    args = cli().parse_args([fn])
    main(args)

def test_ipynb():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.ipynb')
    args = cli().parse_args([fn])
    main(args)

def test_python():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.python')
    args = cli().parse_args([fn])
    with pytest.raises(UnsupportedFileType):
        main(args)

def test_name():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.py')
    args = cli().parse_args([fn, '--name', 'cheese'])
    assert args.name == 'cheese'

def test_max_content_length():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.py')
    args = cli().parse_args([fn, '--max_content_length', '1024'])
    assert args.max_content_length == 1024

def test_port():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.py')
    args = cli().parse_args([fn, '--port', '5000'])
    assert args.port == 5000

def test_address():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.py')
    args = cli().parse_args([fn, '--address', '127.0.0.1'])
    assert args.address == '127.0.0.1'

def test_url_prefix():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.py')
    args = cli().parse_args([fn, '--prefix', 'python'])
    assert args.prefix == 'python'

def test_debug():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.py')
    args = cli().parse_args([fn, '--debug'])
    assert args.debug == True

def test_version():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.py')
    with pytest.raises(SystemExit):
        args = cli().parse_args([fn, '--version'])
        assert args.version == 'tranquilizer {}'.format(__version__)


from tranquilizer import tranquilize

@tranquilize()
def order(cheese):
    '''I'd like to buy some cheese!

    :param cheese: What cheese would you like?'''
    return {'response':"I'm afraid we're fresh out of {}, Sir.".format(cheese)}




from inspect import signature
import re

PARAM_REGEX = re.compile(":param (?P<name>[\*\w]+): (?P<doc>.*?)"
                         "(?:(?=:param)|(?=:return)|(?=:raises)|\Z)", re.S)
RAISE_REGEX = re.compile(":raise[s]? (?P<name>[\*\w]+): (?P<doc>.*?)"
                         "(?:(?=:param)|(?=:return)|(?=:raise[s]?)|\Z)", re.S)

def _prepare_arg(arg):
    '''Return a keyword arg spec (dict)'''
    _arg = {
        "name": arg.name,
    }

    if arg.annotation != arg.empty:
        try:
            name = str(arg.annotation.__name__)
        except AttributeError:
            name = str(arg.annotation._name)
        _arg["type"] = name
        _arg["annotation"] = arg.annotation

    if arg.empty is not arg.default:
        _arg['default'] = arg.default

    return _arg

def _prepare_arg_docs(docstring):
    args = re.findall(PARAM_REGEX, docstring)
    remainder = re.sub(PARAM_REGEX, '', docstring)

    if not args:
        return {}, remainder

    docs = {}
    for arg in args:
        docs[arg[0]] = arg[1].strip()

    return docs, remainder

def _prepare_error_docs(docstring):
    errors = re.findall(RAISE_REGEX, docstring)
    remainder = re.sub(RAISE_REGEX, '', docstring)

    if errors:
        messages = '\n'.join('{}:{}'.format(e,msg.strip()) for e,msg in errors)
        responses = {500:messages}
        return responses, remainder

    return None, remainder

def _prepare(fn):
    """Inspects a function and return a function spec dict

    Output:
        dict containing the following info about f:

        - name: str -> name of the function
        - docstring: str -> docstring of the function
        - args: dict -> dict that holds info about the function arguments.
                    Keys: name of the argument
                    Values: Dict containing the following about the argument:
                        - name: str -> name of the argument
                        - default: any (optional) -> default value
                        - type: any (optional) -> type annotation (if available)

    """
    sig = signature(fn)
    _args = {}
    for k, v in sig.parameters.items():
        _args[k] = _prepare_arg(v)

    if fn.__doc__:
        param_docs, docstring = _prepare_arg_docs(fn.__doc__)
        error_docs, docstring = _prepare_error_docs(docstring)
    else:
        param_docs = {}
        error_docs = {}
        docstring = ''

    spec = {
        'name': fn.__name__,
        'docstring': docstring,
        'args': _args,
        'param_docs': param_docs,
        'error_docs': error_docs
    }

    return spec


def tranquilize(method='get'):
    """Decorator function that gets a function wraps it in order to
    append a function spec (see prepare function) and autocast args/kws
    to match types.

    Parameters
    ----------
    :param method: str, HTTP method for this function. (default: 'get')
    """

    #just to be safe
    method = method.lower()

    def _dart(f):
        f._spec = _prepare(f)
        f._method = method
        f._methods = None
        return f

    return _dart


def publish(methods=['GET']):
    """Decorator function that gets a function wraps it in order to
    append a function spec (see prepare function) and autocast args/kws
    to match types.

    Parameters
    ----------
    :param methods: list, HTTP methods for this function.
                    Provides compatibility with web-publisher.
                    Takes precedence over method. (default: None)
    """

    # just to be safe
    methods = [m.lower() for m in methods]

    def _dart(f):
        f._spec = _prepare(f)
        f._method = None
        f._methods = methods
        return f

    return _dart



from tranquilizer.resource import  make_resource, make_parser
from tranquilizer.handler import ScriptHandler
from tranquilizer.decorator import publish, tranquilize
from tranquilizer.types import is_list_subclass
from flask_restx import Resource, Namespace
import typing
import datetime
import numpy
import PIL.Image
from os.path import dirname, join


def test_make_resource_method():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop.py')
    script = ScriptHandler(fn)
    funcs = script.tranquilized_functions
    ns = Namespace('/', description='Testing Tranquilized API')
    resource = make_resource(funcs[0], ns)
    assert issubclass(resource, Resource)
    assert hasattr(resource, 'get')

def test_make_resource_methods():
    here = dirname(__file__)
    fn = join(here, 'cheese_shop_publish.py')
    script = ScriptHandler(fn)
    funcs = script.tranquilized_functions
    ns = Namespace('/', description='Testing Tranquilized API')
    resource = make_resource(funcs[0], ns)
    assert issubclass(resource, Resource)
    assert hasattr(resource, 'get')
    assert hasattr(resource, 'post')

def test_parser_list():
    def _func(
        l: list,
        L: typing.List,
        Ls: typing.List[str],
        Li: typing.List[int],
        Lf: typing.List[float],
        Lb: typing.List[bool],
        Ld: typing.List[datetime.date],
        Ldt: typing.List[datetime.datetime],
    ):
        pass

    decorated = tranquilize()(_func)
    parser = make_parser(decorated._spec, location='get', compat=False)
    assert len(parser.args) == 8

    assert all([a.action == 'append' for a in parser.args])

def test_parser_files():
    def _func(
        fnb: typing.BinaryIO,
        fnt: typing.TextIO,
        img: PIL.Image.Image,
        arr: numpy.ndarray,
    ):
        pass

    decorated = tranquilize()(_func)
    parser = make_parser(decorated._spec, location='get', compat=False)
    assert len(parser.args) == 4

    assert all([a.location[1] == 'files' for a in parser.args])

def test_parser_default():
    def _func(
        typed_default: str = 'python',
        untyped_default = None,
    ):
        pass

    decorated = tranquilize()(_func)
    parser = make_parser(decorated._spec, location='get', compat=False)
    assert len(parser.args) == 2

    defaults = {a.name:a for a in parser.args}
    assert defaults['untyped_default'].default is None
    assert defaults['typed_default'].default == 'python'

    assert all(a.required == False for a in defaults.values())

def test_parser_builtin():
    def _func(
        s: str,
        i: int,
        f: float,
        b: bool,
        d: datetime.date,
        dt: datetime.datetime,
    ):
        pass

    decorated = tranquilize()(_func)
    parser = make_parser(decorated._spec, location='get', compat=False)
    assert len(parser.args) == 6

    assert all([a.action == 'store' for a in parser.args])

def test_parser_untyped_publish():
    def _func(
        untyped
    ):
        pass

    decorated = publish()(_func)
    parser = make_parser(decorated._spec, location='get', compat=True)
    assert len(parser.args) == 1

    assert parser.args[0].action == 'append'

def test_parser_untyped():
    def _func(
        untyped
    ):
        pass

    decorated = tranquilize()(_func)
    parser = make_parser(decorated._spec, location='get', compat=False)
    assert len(parser.args) == 1

    assert parser.args[0].action == 'store'
    assert parser.args[0].type is str

def test_parser_docs():
    def _func(
        arg: str
    ):
        '''docstring
        
        :param arg: a string'''
        pass

    decorated = tranquilize()(_func)
    parser = make_parser(decorated._spec, location='get', compat=False)
    assert len(parser.args) == 1
    
    assert parser.args[0].help == 'a string'

def test_make_resource_docs():
    here = dirname(__file__)
    fn = join(here, 'integer_error.py')
    script = ScriptHandler(fn)
    funcs = script.tranquilized_functions
    ns = Namespace('/', description='Testing Tranquilized API')
    resource = make_resource(funcs[0], ns)
    assert resource.get.__doc__ == 'Make an integer\n\n    '
    assert resource.get.__apidoc__['responses'] == {500: 'ValueError:not an integer'}



from flask import Flask
from werkzeug.middleware.proxy_fix import ProxyFix
from flask_restx import Api, Namespace

from .resource import make_resource

def make_app(functions, name, prefix='/', max_content_length=None):
    api = Api(title=name)
    app = Flask(__name__)
    app.config['PREFERRED_URL_SCHEME'] = 'https'
    if max_content_length is not None:
        app.config['MAX_CONTENT_LENGTH'] = max_content_length
    app.wsgi_app = ProxyFix(app.wsgi_app,
                            # num_proxies=None,
                            x_for=1, x_proto=1,
                            x_host=1, x_port=1,
                            x_prefix=1)

    @api.errorhandler
    def _default_error(error):
        return {'message': repr(error)}, 500

    ns = Namespace(prefix, description='Tranquilized API')

    for f in functions:
        resource = make_resource(f, ns)
        ns.add_resource(resource, '/{}'.format(f._spec['name']))
        api.add_namespace(ns)

    api.init_app(app)

    return app




from collections.abc import Mapping, Sequence
from dateutil.parser import parse
from datetime import datetime, date
from typing import TextIO, BinaryIO
from werkzeug.datastructures import FileStorage
from flask_restx.reqparse import PY_TYPES
from flask_restx import fields
import io

def is_container(type_):
    '''Test if provided type function is a scalar

    strings and bytes are considered to be scalars.'''

    origin = getattr(type_, '__origin__', None)
    if origin is not None:
        # Due to PEP560 we need to handle new typing
        # classes carefully
        return issubclass(type_.__origin__, Sequence)

    container = issubclass(type_, Sequence) or (issubclass(type_, Mapping))
    basic_scalars = issubclass(type_, str) or issubclass(type_, bytes)

    return (not basic_scalars) and container


def is_list_subclass(type_):
    origin = getattr(type_, '__origin__', None)
    if origin is not None:
        return issubclass(type_.__origin__, list)
    else:
        return issubclass(type_, list)


class File(FileStorage):
    '''Thin wrapper for werkzeug.datastructures.FileStorage'''
    __location__ = 'files'
    __schema__ = {'type':'file', 'format':'binary file'}
    __description__ = 'Read file as binary.'
    def __new__(cls, file):
        return file


class TextFile(File):
    '''Returns io.StringIO object'''
    __schema__ = {'type':'file', 'format':'text file'}
    __description__ = 'Read file as text.'
    def __new__(cls, *args, **kwargs):
        f = super().__new__(cls, *args, **kwargs)
        t = io.StringIO(f.read().decode())
        return t


class Image(File):
    '''Returns PIL.Image object'''
    __schema__ = {'type':'file', 'format':'image'}
    __description__ = 'Read image file as PIL.Image.'
    def __new__(cls, *args, **kwargs):
        try:
            from PIL import Image as pil_image
        except ImportError as e: # pragma: no cover
            e.args = ("Please install pillow to define an image type.",)
            raise e
        f = super().__new__(cls, *args, **kwargs)

        return pil_image.open(f)


class NDArray(File):
    '''Returns a NumPy array using np.load()'''
    __schema__ = {'type':'file', 'format':'NumPy array'}
    __description__ = 'NumPy array file.'
    def __new__(cls, *args, **kwargs):
        try:
            import numpy as np
        except ImportError as e: # pragma: no cover
            e.args = ("Please install NumPy to define an array type.",)
            raise e
        f = super().__new__(cls, *args, **kwargs)
        return np.load(f)

def dt_factory(type_):
    class ParsedDatetime(type):
        '''A flexible datetime class

        receives a string returns datetime object
        '''
        __schema__ = {'type': 'string'}
        __description__ = 'dateutil.parser.parse compatible datetime string'
        def __new__(self, arg):
            if type_ is datetime:
                return parse(arg)
            elif type_ is date:
                return parse(arg).date()
            else:
                # most commonly used for pd.TimeStamp.
                # any method that can take a string
                return type_(arg)

    return ParsedDatetime


def list_factory(type_):
    items = getattr(type_, '__schema__', {}).get('type', PY_TYPES.get(type_, 'string'))

    class TypedList(Sequence):
        # using append with flask-restx
        # means that list is constructed later
        __schema__ = {'type': items}
        __description__ = 'List with values of type {}'.format(items)
        def __new__(cls, arg):
            return type_(arg)
        
    return TypedList


def type_mapper(type_):
    '''Map common type hints to custom classes

    If no conversion is necessary the input type
    is returned.
    '''

    try:
        from PIL import Image as pil_image
        has_pil_image = True
    except ImportError: # pragma: no cover 
        has_pil_image = False

    try:
        import numpy as np
        has_numpy = True
    except ImportError: # pragma: no cover 
        has_numpy = False

    if is_list_subclass(type_):
        try:
            item_type = type_.__args__[0]
            if issubclass(item_type, (datetime, date)):
                item_type = dt_factory(item_type)
        except (AttributeError, TypeError):
            item_type = str
        return list_factory(item_type)
    elif issubclass(type_, TextIO):
        return TextFile
    elif issubclass(type_, BinaryIO):
        return File
    elif issubclass(type_, (datetime, date)):
        return dt_factory(type_)
    elif has_pil_image and issubclass(type_, pil_image.Image):
        return Image
    elif has_numpy and issubclass(type_, np.ndarray):
        return NDArray
    else:
        return type_
